@@ -20,18 +20,20 @@ If you want to use this module instead of autoloading another event loop
 you can simply load it before creating the first watcher.
 
 As for performance, this module is on par with (and usually faster than)
-most select/poll-based C event modules such as Event or Glib (it does
-not come close to EV, though), with respect to I/O watchers. Timers are
-handled less optimally.
+most select/poll-based C event modules such as Event or Glib (it does not
+even come close to EV, though), with respect to I/O watchers. Timers are
+handled less optimally, but for many common tasks, it's still on par with
+event loops written in C.
 
-This event loop has been optimised for the following cases:
+This event loop has been optimised for the following use cases:
 
 =over 4
 
 =item monotonic clock is available
 
-This module will use the POSIX monotonic clock option if it can be
-detected at runtime, in which case it will not suffer adversely from time
+This module will use the POSIX monotonic clock option (if it can be
+detected at runtime) or the POSIX C<times> function (if the resolution is
+better than 100Hz), in which case it will not suffer adversely from time
 jumps.
 
 If no monotonic clock is available, this module will not attempt to
@@ -40,29 +42,40 @@ correct for time jumps in any way.
 The clock chosen will be reported if the environment variable
 C<$PERL_ANYEVENT_VERBOSE> is set to 8 or higher.
 
-=item lots of watchers on one fd
+=item any number of watchers on one fd
 
-This is purely a dirty benchmark optimisation not relevant in practise.
-The more common case of having one watcher per fd/poll combo is
-special-cased and still fast.
+Supporting a large number of watchers per fd is purely a dirty benchmark
+optimisation not relevant in practise. The more common case of having one
+watcher per fd/poll combo is special-cased, however, and therefore fast,
+too.
 
-=item relatively few active fds per select call
+=item relatively few active fds per C<select> call
 
 This module expects that only a tiny amount of fds is active at any one
 time. This is relatively typical of larger servers (but not the case where
-select traditionally is fast), at the expense of the "dense activity case"
-where most of fds are active (which suits select and poll). The optimal
-implementation of the "dense" case is not much faster, though, so the
-module should behave very well in most cases.
+C<select> traditionally is fast), at the expense of the "dense activity
+case" where most of the fds are active (which suits C<select>).
+
+The optimal implementation of the "dense" case is not much faster, though,
+so the module should behave very well in most cases, subject to the bad
+scalability of C<select> in general.
 
 =item lots of timer changes/iteration, or none at all
 
-This module sorts the timer list again on each iteration, if timers have
-been created. If you add lots of timers, this is pretty efficient, but the
-common case of only adding a few timers is not handled very efficiently.
+This module sorts the timer list using perl's C<sort>, even though a total
+ordering is not required for timers.
 
-This should not have much of an impact unless you have hundreds of timers,
-though.
+This sorting is expensive, but means sorting can be avoided unless the
+timer list has changed in a way that requires a new sort.
+
+This means that adding lots of timers is very efficient, as well as not
+changing the timers. Advancing timers (e.g. recreating a timeout watcher
+on activity) is also relatively efficient, for example, if you have a
+large number of timeout watchers that time out after 10 seconds, then the
+timer list will be sorted only once every 10 seconds.
+
+This should not have much of an impact unless you have hundreds or
+thousands of timers, though, or your timers have very small timeouts.
 
 =back
 
@@ -71,33 +84,63 @@ though.
 package AnyEvent::Impl::Perl;
 
 no warnings;
-use strict;
+use strict qw(vars subs);
 
-use Time::HiRes ();
-use Scalar::Util ();
+use Scalar::Util qw(weaken);
 
 use AnyEvent ();
 use AnyEvent::Util ();
 
-our $VERSION = 4.22;
+our $VERSION = 4.331;
 
 our ($NOW, $MNOW);
 
+sub MAXWAIT() { 1000 } # never sleep for longer than this many seconds
+
 BEGIN {
    local $SIG{__DIE__};
+   my $time_hires = eval "use Time::HiRes (); 1";
+   my $clk_tck    = eval "use POSIX (); POSIX::sysconf (&POSIX::_SC_CLK_TCK)";
+   my $round; # actual granularity
 
-   if (0 < eval "&Time::HiRes::clock_gettime (&Time::HiRes::CLOCK_MONOTONIC)") {
+   if ($time_hires && eval "&Time::HiRes::clock_gettime (&Time::HiRes::CLOCK_MONOTONIC)") {
+      warn "AnyEvent::Impl::Perl using CLOCK_MONOTONIC as timebase.\n" if $AnyEvent::verbose >= 8;
       *_update_clock = sub {
-         $NOW  = Time::HiRes::time;
+         $NOW  = &Time::HiRes::time;
          $MNOW = Time::HiRes::clock_gettime (&Time::HiRes::CLOCK_MONOTONIC);
       };
-      warn "AnyEvent::Impl::Perl using CLOCK_MONOTONIC as timebase.\n" if $AnyEvent::verbose >= 8;
-   } else {
+
+   } elsif (100 <= $clk_tck && $clk_tck <= 1000000 && "use POSIX (); (POSIX::times())[0] != -1") { # -1 is also a valid return value :/
+      warn "AnyEvent::Impl::Perl using POSIX::times (monotonic) as timebase.\n" if $AnyEvent::verbose >= 8;
+      my $HZ1 = 1 / $clk_tck;
+
+      my $last = (POSIX::times ())[0];
+      my $next;
       *_update_clock = sub {
-         $NOW = $MNOW = Time::HiRes::time;
+         $NOW  = time; # d'oh
+
+         $next = (POSIX::times ())[0];
+         # we assume 32 bit signed on wrap but 64 bit will never wrap
+         $last -= 0x100000000 if $last > $next;
+         $MNOW += ($next - $last) * $HZ1;
+         $last = $next;
       };
-      warn "AnyEvent::Impl::Perl using default (non-monotonic) clock as timebase.\n" if $AnyEvent::verbose >= 8;
+
+      $round = $HZ1;
+
+   } elsif (eval "use Time::HiRes (); 1") {
+      warn "AnyEvent::Impl::Perl using Time::HiRes::time (non-monotonic) clock as timebase.\n" if $AnyEvent::verbose >= 8;
+      *_update_clock = sub {
+         $NOW = $MNOW = &Time::HiRes::time;
+      };
+
+   } else {
+      die "FATAL: unable to find sub-second time source (is this really perl 5.8.0 or later?)";
    }
+
+   $round = 0.001 if $round < 0.001; # 1ms is enough for us
+   $round -= $round * 1e-2; # 0.1 => 0.099
+   eval "sub ROUNDUP() { $round }";
 }
 
 _update_clock;
@@ -105,12 +148,14 @@ _update_clock;
 sub now { $NOW }
 
 # fds[0] is for read, fds[1] is for write watchers
-# fds[poll]{v} is the bitmask for select
-# fds[poll]{w}[fd] contains a list of i/o watchers
+# fds[poll][V] is the bitmask for select
+# fds[poll][W][fd] contains a list of i/o watchers
 # an I/O watcher is a blessed arrayref containing [fh, poll(0/1), callback, queue-index]
-# the queue-index is simply the index in the {w} array, which is only used to improve
+# the queue-index is simply the index in the [W] array, which is only used to improve
 # benchmark results in the synthetic "many watchers on one fd" benchmark.
-my @fds = ({}, {});
+my @fds = ([], []);
+sub V() { 0 }
+sub W() { 1 }
 
 my $need_sort = 1e300; # when to re-sort timer list
 my @timer; # list of [ abs-timeout, Timer::[callback] ]
@@ -139,11 +184,13 @@ sub one_event {
       my ($wait, @vec, $fds)
          = (@timer && $timer[0][0] < $need_sort ? $timer[0][0] : $need_sort) - $MNOW;
 
+      $wait = $wait < MAXWAIT ? $wait + ROUNDUP : MAXWAIT;
+
       if ($fds = select
-            $vec[0] = $fds[0]{v},
-            $vec[1] = $fds[1]{v},
-            AnyEvent::WIN32 ? $vec[2] = $fds[1]{v} : undef,
-            $wait < 3600 ? $wait + 0.0009 : 3600
+            $vec[0] = $fds[0][V],
+            $vec[1] = $fds[1][V],
+            AnyEvent::WIN32 ? $vec[2] = $fds[1][V] : undef,
+            $wait
       ) {
          _update_clock;
 
@@ -162,13 +209,13 @@ sub one_event {
                while (/\G0*1/g) {
                   # and using the resulting string position as fd
                   $_ && $_->[2]()
-                     for @{ $fds->{w}[-1 + pos] || [] };
+                     for @{ $fds->[W][-1 + pos] || [] };
                }
             }
          }
       } elsif (AnyEvent::WIN32 && $! == AnyEvent::Util::WSAEINVAL) {
          # buggy microshit windoze asks us to route around it
-         select undef, undef, undef, @timer ? $timer[0][0] - $MNOW  + 0.0009 : 3600;
+         select undef, undef, undef, $wait;
       }
    }
 }
@@ -176,8 +223,10 @@ sub one_event {
 sub io {
    my ($class, %arg) = @_;
 
+   my $fd = fileno $arg{fh};
+
    my $self = bless [
-      $arg{fh},
+      $fd,
       $arg{poll} eq "w",
       $arg{cb},
       # q-idx
@@ -186,14 +235,13 @@ sub io {
    my $fds = $fds[$self->[1]];
 
    # add watcher to fds structure
-   my $fd = fileno $self->[0];
-   my $q = $fds->{w}[$fd] ||= [];
+   my $q = $fds->[W][$fd] ||= [];
 
-   (vec $fds->{v}, $fd, 1) = 1;
+   (vec $fds->[V], $fd, 1) = 1;
 
    $self->[3] = @$q;
    push @$q, $self;
-   Scalar::Util::weaken $q->[-1];
+   weaken $q->[-1];
 
    $self
 }
@@ -204,17 +252,17 @@ sub AnyEvent::Impl::Perl::Io::DESTROY {
    my $fds = $fds[$self->[1]];
 
    # remove watcher from fds structure
-   my $fd = fileno $self->[0];
+   my $fd = $self->[0];
 
-   if (@{ $fds->{w}[$fd] } == 1) {
-      delete $fds->{w}[$fd];
-      (vec $fds->{v}, $fd, 1) = 0;
+   if (@{ $fds->[W][$fd] } == 1) {
+      delete $fds->[W][$fd];
+      (vec $fds->[V], $fd, 1) = 0;
    } else {
-      my $q = $fds->{w}[$fd];
+      my $q = $fds->[W][$fd];
       my $last = pop @$q;
 
       if ($last != $self) {
-         Scalar::Util::weaken ($q->[$self->[3]] = $last);
+         weaken ($q->[$self->[3]] = $last);
          $last->[3] = $self->[3];
       }
    }
@@ -232,7 +280,7 @@ sub timer {
       $self = [$MNOW + $arg{after} , sub {
          $_[0][0] = $MNOW + $ival;
          push @timer, $_[0];
-         Scalar::Util::weaken $timer[-1];
+         weaken $timer[-1];
          $need_sort = $_[0][0] if $_[0][0] < $need_sort;
          &$cb;
       }];
@@ -241,7 +289,7 @@ sub timer {
    }
 
    push @timer, $self;
-   Scalar::Util::weaken $timer[-1];
+   weaken $timer[-1];
    $need_sort = $self->[0] if $self->[0] < $need_sort;
 
    $self
